""" Token Definition """
from rowantree.contracts import BaseModel


class Token(BaseModel):
    """
    Token

    Attributes
    ----------
    access_token: str
        Minted access token
    token_type: str
        Type of token
    """

    access_token: str
    token_type: str



""" Register User Command Definition """

from starlette import status

from ..contracts.dto.request_status_codes import RequestStatusCodes
from ..contracts.dto.user.base import UserBase
from ..contracts.dto.wrapped_request import WrappedRequest
from ..contracts.request_verb import RequestVerb
from ..contracts.requests.register_user_request import RegisterUserRequest
from .abstract_command import AbstractCommand


class RegisterUserCommand(AbstractCommand):
    """
    Register User Command

    Methods
    -------
    execute(self, request: User) -> None
        Register User
    """

    def execute(self, request: RegisterUserRequest) -> UserBase:
        """
        Register User

        Parameters
        ----------
        request: RegisterUserRequest
        """

        request: WrappedRequest = WrappedRequest(
            verb=RequestVerb.FORM,
            url=f"https://api.{self.options.tld}/auth/v1/register",
            statuses=RequestStatusCodes(allow=[status.HTTP_200_OK], reauth=[], retry=[]),
            data=request.dict(by_alias=True),
        )
        response: dict = self.wrapped_request(request=request)
        return UserBase.parse_obj(response)



""" Authenticate User Request Definition """
from rowantree.contracts import BaseModel


class AuthenticateUserRequest(BaseModel):
    """
    Authenticate User Request

    Attributes
    ----------
    username: str
        Username
    password: str
        Plain text password
    """

    username: str
    password: str



""" User Definition """

from .base import UserBase


class User(UserBase):
    """
    User Definition

    Attributes
    ----------
    hashed_password: str
        Hashed user password.
    """

    hashed_password: str



""" Password Related Utilities """

from passlib.context import CryptContext

# Acts as a singleton
ROWANTREE_AUTH_SDK_PASSWORD_CONTEXT: CryptContext = CryptContext(schemes=["bcrypt"], deprecated="auto")


def verify_password(plain_password: str, hashed_password: str) -> bool:
    """
    Verifies hashed password.

    Parameters
    ----------
    plain_password: str
        Plan text password
    hashed_password: str
        Hashed password

    Returns
    -------
    verified: bool
        True if the password matches the hash, False otherwise.
    """

    return ROWANTREE_AUTH_SDK_PASSWORD_CONTEXT.verify(plain_password, hashed_password)


def get_password_hash(password: str) -> str:
    """
    Generates password hash.

    Parameters
    ----------
    password: str
        Plain text password.

    Returns
    -------
    hashed_password: str
        Hashed password.
    """

    return ROWANTREE_AUTH_SDK_PASSWORD_CONTEXT.hash(password)



""" Action Queue Process Command Definition """

from starlette import status

from ..contracts.dto.request_status_codes import RequestStatusCodes
from ..contracts.dto.token import Token
from ..contracts.dto.wrapped_request import WrappedRequest
from ..contracts.request_verb import RequestVerb
from ..contracts.requests.authenticate_user_request import AuthenticateUserRequest
from .abstract_command import AbstractCommand


class AuthenticateUserCommand(AbstractCommand):
    """
    Authenticate User Command

    Methods
    -------
    execute(self, request: AuthenticateUserRequest) -> None
        Executes the command.
    """

    def execute(self, request: AuthenticateUserRequest) -> Token:
        """
        Executes the command.

        Parameters
        ----------
        request: AuthenticateUserRequest
        """

        request: WrappedRequest = WrappedRequest(
            verb=RequestVerb.FORM,
            url=f"https://api.{self.options.tld}/auth/v1/token",
            statuses=RequestStatusCodes(allow=[status.HTTP_200_OK], reauth=[status.HTTP_401_UNAUTHORIZED], retry=[]),
            data=request.dict(by_alias=True),
        )
        response: dict = self.wrapped_request(request=request)
        return Token.parse_obj(response)



""" rowantree.auth.sdk namespace """

from .commands.abstract_command import AbstractCommand
from .commands.authenticate_user import AuthenticateUserCommand
from .commands.register_user import RegisterUserCommand
from .common.claims import get_claims, get_issuer
from .common.depends import get_current_user, is_admin, is_enabled, oauth2_scheme
from .common.jwt import create_access_token, create_user_access_token
from .common.password import get_password_hash, verify_password
from .contracts.dto.command_options import CommandOptions
from .contracts.dto.request_status_codes import RequestStatusCodes
from .contracts.dto.token import Token
from .contracts.dto.token_claims import TokenClaims
from .contracts.dto.user.base import UserBase
from .contracts.dto.user.user import User
from .contracts.dto.wrapped_request import WrappedRequest
from .contracts.exceeded_retry_count_error import ExceededRetryCountError
from .contracts.request_failure_error import RequestFailureError
from .contracts.request_verb import RequestVerb
from .contracts.requests.authenticate_user_request import AuthenticateUserRequest
from .contracts.requests.register_user_request import RegisterUserRequest



""" Request Verb Definition """

from enum import Enum


class RequestVerb(str, Enum):
    """Defines the types of requests supported."""

    GET = "GET"
    POST = "POST"
    FORM = "FORM"
    DELETE = "DELETE"



""" Auth Service Register User Base Definition """
from rowantree.contracts import BaseModel


class RegisterUserRequest(BaseModel):
    """
    Auth Service Register User Base

    Attributes
    ----------
    username: str
        The username of the user.
    email: Union[str, None] = None
        The user email address.
    password: str
        user password.
    """

    username: str
    email: str
    password: str



""" Claim Utilities """

import logging
from typing import Optional

from jose import JWTError, jwt
from starlette import status
from starlette.exceptions import HTTPException

from rowantree.common.sdk import demand_env_var

from ..contracts.dto.token_claims import TokenClaims


def get_claims(token: str) -> TokenClaims:
    """
    Gets claims from an OAuth2 token.

    Parameters
    ----------
    token: str
        Encoded token.

    Returns
    -------
    claims: TokenClaims
        An instance of TokenClaims, otherwise an HTTPException exception.
    """

    credentials_exception: HTTPException = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )

    try:
        issuer: str = get_issuer()

        payload: dict = jwt.decode(
            token=token,
            key=demand_env_var(name="ACCESS_TOKEN_SECRET_KEY"),
            algorithms=[demand_env_var(name="ACCESS_TOKEN_ALGORITHM")],
            issuer=issuer,
        )
        token_issuer: Optional[str] = payload.get("iss")
        guid: Optional[str] = payload.get("sub")
        if token_issuer != issuer or guid is None:
            logging.debug("Received issuer: %s, expected: %s, guid was: %s", token_issuer, issuer, guid)
            raise credentials_exception
        logging.debug(payload)
        return TokenClaims(**payload)
    except JWTError:
        raise credentials_exception from JWTError


def get_issuer() -> str:
    """
    Builds issuer string

    Returns
    -------
    issuer: str
        The issuer for the domain.
    """

    return f"https://api.{demand_env_var(name='ROWANTREE_TLD')}/auth"



""" Defines the route level auth dependencies for use with FastAPI """

from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer
from starlette.exceptions import HTTPException

from ..contracts.dto.token_claims import TokenClaims
from .claims import get_claims, get_issuer

oauth2_scheme = OAuth2PasswordBearer(tokenUrl=get_issuer())


def get_current_user(token: str = Depends(oauth2_scheme)) -> TokenClaims:
    """
    Gets token claims for token.

    Parameters
    ----------
    token: str
        Oauth2 Token

    Returns
    -------
    claims: TokenClaims
        An instance of TokenClaims.
    """

    return get_claims(token=token)


def is_enabled(token_claims: TokenClaims = Depends(get_current_user)) -> TokenClaims:
    """
    Enforces that the user is enabled.

    Parameters
    ----------
    token_claims: TokenClaims
        User claims to review.

    Returns
    -------
    claims: TokenClaims
        An instance of TokenClaims, otherwise an HTTPException exception.
    """

    if token_claims.disabled:
        raise HTTPException(status_code=401, detail="Inactive user")
    return token_claims


def is_admin(token_claims: TokenClaims = Depends(is_enabled)) -> TokenClaims:
    """
    Enforces that the user is an admin.

    Parameters
    ----------
    token_claims: TokenClaims
        User claims to review.

    Returns
    -------
    claims: TokenClaims
        An instance of TokenClaims, otherwise an HTTPException exception.
    """

    if not token_claims.admin:
        raise HTTPException(status_code=401, detail="Insufficient Permissions")
    return token_claims



""" Abstract Command Definition """
import json
import logging
import time
from abc import abstractmethod
from typing import Any, Optional

import requests
from requests import Response

from rowantree.common.sdk import demand_env_var, demand_env_var_as_float, demand_env_var_as_int
from rowantree.contracts import BaseModel

from ..contracts.dto.command_options import CommandOptions
from ..contracts.dto.wrapped_request import WrappedRequest
from ..contracts.exceeded_retry_count_error import ExceededRetryCountError
from ..contracts.request_failure_error import RequestFailureError
from ..contracts.request_verb import RequestVerb


class AbstractCommand(BaseModel):
    """
    Abstract Command
    """

    options: Optional[CommandOptions]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.options is None:
            self.options = CommandOptions(
                sleep_time=demand_env_var_as_float(name="ACCESS_AUTH_ENDPOINT_SLEEP_TIME"),
                retry_count=demand_env_var_as_int(name="ACCESS_AUTH_ENDPOINT_RETRY_COUNT"),
                tld=demand_env_var(name="ROWANTREE_TLD"),
                timeout=demand_env_var_as_float(name="ACCESS_AUTH_ENDPOINT_TIMEOUT"),
            )

    @abstractmethod
    def execute(self, *args, **kwargs) -> Optional[Any]:
        """Command entry point"""

    def _build_requests_params(self, request: WrappedRequest) -> dict:
        """
        Builds the `requests` call parameters.

        Parameters
        ----------
        request: WrappedRequest
            The request to make.

        Returns
        -------
        params: dict
            A dictionary suitable for splatting into the `requests` call.
        """

        params: dict = {
            "url": request.url,
            "timeout": self.options.timeout,
        }
        if request.verb == RequestVerb.POST:
            params["data"] = request.data
        if request.verb == RequestVerb.FORM:
            params["files"] = {}
            for key in request.data.keys():
                params["files"][key] = (None, request.data[key])
        if request.params is not None:
            params["params"] = request.params
        return params

    def _api_caller(self, request: WrappedRequest, depth: int) -> dict:
        """
        Wrapper for calls with `requests` to external APIs.

        Parameters
        ----------
        request: WrappedRequest
            Request to make.
        depth: int
            Call depth of the recursive call (retry)

        Returns
        -------
        response: dict
            A dictionary of the response.
        """

        if depth < 1:
            raise ExceededRetryCountError(json.dumps({"request": request.dict(by_alias=True), "depth": depth}))
        depth -= 1

        params: dict = self._build_requests_params(request=request)
        # pylint: disable=broad-except
        try:
            if request.verb == RequestVerb.GET:
                response: Response = requests.get(**params)
            elif request.verb == RequestVerb.POST:
                response: Response = requests.post(**params)
            elif request.verb == RequestVerb.DELETE:
                response: Response = requests.delete(**params)
            elif request.verb == RequestVerb.FORM:
                response: Response = requests.post(**params)
            else:
                raise Exception("Unknown Verb")
        except requests.exceptions.ConnectionError as error:
            logging.debug("Connection Error (%s) - Retrying.. %i", str(error), depth)
            time.sleep(self.options.sleep_time)
            return self._api_caller(request=request, depth=depth)
        except Exception as error:
            logging.debug("Exception needed to cover: %s", str(error))
            time.sleep(self.options.sleep_time)
            return self._api_caller(request=request, depth=depth)

        if response.status_code in request.statuses.allow:
            return response.json()

        if response.status_code in request.statuses.retry:
            time.sleep(self.options.sleep_time)
            return self._api_caller(request=request, depth=depth)

        raise RequestFailureError(
            json.dumps({"status_code": response.status_code, "request": request.dict(by_alias=True), "depth": depth})
        )

    def wrapped_request(self, request: WrappedRequest) -> dict:
        """
        High level request method.  Entry point for consumption.


        Parameters
        ----------
        request: WrappedRequest
            The request to make.

        Returns
        -------
        response: dict
            The response as a dictionary.
        """

        return self._api_caller(request=request, depth=self.options.retry_count)



""" Wrapped Request Definition """

from typing import Optional

from rowantree.contracts import BaseModel

from ...contracts.request_verb import RequestVerb
from .request_status_codes import RequestStatusCodes


class WrappedRequest(BaseModel):
    """
    Wrapped Request DTO

    Attributes
    ----------
    verb: RequestVerb
        The verb of the REST API request.
    statuses: RequestStatusCodes
        The status code assignments for response handling.
    url: str
        The complete url to call
    data: Optional[Any]  # str or dict ? - needs confirmation
        Either form data or body (based on type - see requests documentation)
    params: Optional[dict[str, str]]
        Query parameters
    """

    verb: RequestVerb
    statuses: RequestStatusCodes
    url: str
    data: Optional[dict]
    params: Optional[dict[str, str]]



""" Token Claims Definition """

from datetime import datetime

from pydantic import BaseModel


class TokenClaims(BaseModel):
    """
    Token Claims

    Attributes
    ----------
    sub: str
        The subject of the token. (the target user guid)
    iss: str
        The issuer of the token (the auth url for the current domain)
    exp: datetime
        The expiry time of the token
    disabled: bool
        Whether or not the user is disabled
    admin: bool
        Whether or not the user is ad admin.
    """

    sub: str
    iss: str
    exp: datetime

    # TOO: move to scopes
    disabled: bool
    admin: bool



""" JWT Related Utilities """

from datetime import datetime, timedelta

from jose import jwt

from rowantree.common.sdk import demand_env_var, demand_env_var_as_float

from ..contracts.dto.token import Token
from ..contracts.dto.user.user import User
from .claims import get_issuer


def create_user_access_token(user: User) -> Token:
    """
    Create bearer Token from user instance.

    Parameters
    ----------
    user: User
        An instance of a UserInDB object.

    Returns
    -------
    token: Token
        A bearer token for the requested user.
    """

    # Currently these are the claims we pull from the database.
    data: dict = {"sub": user.guid, "disabled": user.disabled, "admin": user.admin}
    return create_access_token(data=data)


def create_access_token(data: dict) -> Token:
    """
    Mints the access token.

    Parameters
    ----------
    data: dict
        The base set of claims to include in the token.

    Returns
    -------
    token: Token
        A newly minted token.
    """

    to_encode: dict = data.copy()
    expire: datetime = datetime.utcnow() + timedelta(
        minutes=demand_env_var_as_float(name="ACCESS_TOKEN_EXPIRATION_TIME")
    )
    to_encode.update({"iss": get_issuer(), "exp": expire})
    encoded_jwt: str = jwt.encode(
        to_encode,
        demand_env_var(name="ACCESS_TOKEN_SECRET_KEY"),
        algorithm=demand_env_var(name="ACCESS_TOKEN_ALGORITHM"),
    )
    return Token(access_token=encoded_jwt, token_type="bearer")
